import asyncio
import collections
import logging

from irc.client_aio import AioConnection
from irc.client_aio import AioReactor
from irc.client_aio import IrcProtocol
from irc.connection import AioFactory


class MultiQueue:
    def __init__(self):
        self._prios = []
        self._ques = {}

    def __len__(self):
        return sum([len(q) for q in self._ques.values()])

    def append(self, item):
        prio, value, tag = item

        if prio not in self._prios:
            self._prios.append(prio)
            self._prios.sort()
            self._ques[prio] = collections.deque()

        self._ques[prio].append(item)

    def get(self):
        for prio in self._prios:
            que = self._ques[prio]
            if len(que) > 0:
                return que.popleft()

        raise IndexError("Get called when all queues empty")

    def filter(self, func) -> int:
        filtered = 0

        for que in self._ques.values():
            tmp = que.copy()
            olen = len(que)
            que.clear()
            que.extend(filter(func, tmp))
            filtered += olen - len(que)

        return filtered


# asyncio.PriorityQueue does not preserve order within priority level
class OrderedPriorityQueue(asyncio.Queue):
    def _init(self, maxsize):
        self._queue = MultiQueue()

    def _get(self):
        return self._queue.get()

    def _put(self, item):
        self._queue.append(item)

    def remove_tag(self, tag) -> int:
        return self._queue.filter(lambda x: x == tag)


class HeisenProtocol(IrcProtocol):
    ping_timeout = 300

    def connection_made(self, *args, **kwargs):
        super().connection_made(*args, **kwargs)

        # start aliveness check
        self._timer = self.loop.call_later(60, self._are_we_still_alive)
        self._last_data = self.loop.time()

    def connection_lost(self, exc):
        super().connection_lost(exc)
        self._timer.cancel()

    def data_received(self, *args, **kwargs):
        super().data_received(*args, **kwargs)
        self._last_data = self.loop.time()

    def _are_we_still_alive(self):
        if not self.connection or not hasattr(self.connection, "connected") or not self.connection.connected:
            return

        # no
        if self.loop.time() - self._last_data >= self.ping_timeout:
            logging.debug("Disconnecting due to no data received from server.")
            self.connection.disconnect("No data received.")
            return

        # re-schedule aliveness check
        self._timer = self.loop.call_later(self.ping_timeout / 3, self._are_we_still_alive)

        # yes
        if self.loop.time() - self._last_data < self.ping_timeout / 3:
            return

        # perhaps, ask the server
        logging.debug("Aliveness check failed, sending PING")
        self.connection.send_items("PING", self.connection.real_server_name)


class HeisenConnection(AioConnection):
    protocol_class = HeisenProtocol

    def __init__(self, reactor):
        super().__init__(reactor)
        self._queue = OrderedPriorityQueue()

    async def expect(self, events, timeout=30):
        events = events if not isinstance(events, str) and not isinstance(events, int) else [events]
        waitable = asyncio.Event()
        result = None

        def expected(connection, event):
            nonlocal result, waitable
            result = (connection, event)
            waitable.set()
            return "NO MORE"

        for event in events:
            self.add_global_handler(event, expected, -100)

        try:
            await asyncio.wait_for(waitable.wait(), timeout)
            return result
        finally:
            for event in events:
                self.remove_global_handler(event, expected)

    async def connect(
        self,
        server,
        port,
        nickname,
        password=None,
        username=None,
        ircname=None,
        connect_factory=AioFactory(),
    ):
        if self.connected:
            self.disconnect("Changing servers")

        self.buffer = self.buffer_class()
        self.handlers = {}
        self.real_server_name = ""
        self.real_nickname = ""
        self.server = server
        self.port = port
        self.server_address = (server, port)
        self.nickname = nickname
        self.username = username or nickname
        self.ircname = ircname or nickname
        self.password = password
        self.connect_factory = connect_factory

        protocol_instance = self.protocol_class(self, self.reactor.loop)
        connection = self.connect_factory(protocol_instance, self.server_address)
        transport, protocol = await connection

        self.transport = transport
        self.protocol = protocol

        self.connected = True
        self._task = asyncio.ensure_future(self._run())
        self.reactor._on_connect(self.protocol, self.transport)
        return self

    async def register(self):
        # Log on...
        if self.password:
            self.pass_(self.password)
        self.nick(self.nickname)
        self.user(self.username, self.ircname)

    def close(self):
        logging.debug("Canceling IRC event queue")
        self._task.cancel()
        super().close()

    async def _run(self):
        loop = asyncio.get_running_loop()
        last = loop.time()
        penalty = 0

        while True:
            try:
                (priority, string, tag) = await self._queue.get()

                diff = int(loop.time() - last)

                # zero int diff means we are going too fast
                if diff == 0:
                    penalty += 1
                else:
                    penalty -= diff
                    if penalty < 0:
                        penalty = 0

                super().send_raw(string)

                # sleep is based on message length
                sleep_time = max(len(string.encode()) / 512 * 6, 1.5)

                if penalty > 5 or sleep_time > 1.5:
                    await asyncio.sleep(sleep_time)

                # this needs to be reset if we slept
                last = loop.time()
            except asyncio.CancelledError:
                break
            except Exception:
                logging.exception("Failed to flush IRC queue")

            self._queue.task_done()

        logging.debug("IRC event queue ended")

    def send_raw(self, string, priority=0, tag=None):
        self._queue.put_nowait((priority, string, tag))

    def send_items(self, *items):
        priority = 0
        tag = None
        if items[0] == "NOTICE":
            # queue CTCP replies even lower than notices
            if len(items) > 2 and len(items[2]) > 1 and items[2][1] == "\001":
                priority = 3
            else:
                priority = 2
        if items[0] == "PRIVMSG":
            priority = 1
        elif items[0] == "PONG":
            priority = -1

        # tag with target to dequeue with filter
        if tag is None and items[0] in ["NOTICE", "PRIVMSG", "MODE", "JOIN", "PART", "KICK"]:
            tag = items[1].lower()

        self.send_raw(" ".join(filter(None, items)), priority, tag)

    def remove_tag(self, tag) -> int:
        return self._queue.remove_tag(tag)


class HeisenReactor(AioReactor):
    connection_class = HeisenConnection

    def _handle_event(self, connection, event):
        with self.mutex:
            matching_handlers = sorted(self.handlers.get("all_events", []) + self.handlers.get(event.type, []))

            if len(matching_handlers) == 0 and event.type != "all_raw_messages" and event.type != "pong":
                matching_handlers += self.handlers.get("unhandled_events", [])

            for handler in matching_handlers:
                result = handler.callback(connection, event)
                if result == "NO MORE":
                    return



import argparse
import shlex


class CommandParserFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter):
    pass


class CommandParserError(Exception):
    pass


class CommandParser(argparse.ArgumentParser):
    def __init__(self, *args, formatter_class=CommandParserFormatter, **kwargs):
        super().__init__(*args, formatter_class=formatter_class, **kwargs)

    @property
    def short_description(self):
        return self.description.split("\n")[0]

    def error(self, message):
        raise CommandParserError(message)

    def print_usage(self):
        raise CommandParserError(self.format_usage())

    def print_help(self):
        raise CommandParserError(self.format_help())

    def exit(self, status=0, message=None):
        pass


def split(text):
    commands = []

    sh_split = shlex.shlex(text, posix=True, punctuation_chars=";")
    sh_split.commenters = ""
    sh_split.wordchars += "!#$%&()*+,-./:<=>?@[\\]^_`{|}~"

    args = []
    for v in list(sh_split):
        if v == ";":
            commands.append(args)
            args = []
        else:
            args.append(v)

    if len(args) > 0:
        commands.append(args)

    return commands


class CommandManager:
    _commands: dict

    def __init__(self):
        self._commands = {}

    def register(self, cmd: CommandParser, func, aliases=None):
        self._commands[cmd.prog] = (cmd, func)

        if aliases is not None:
            for alias in aliases:
                self._commands[alias] = (cmd, func)

    async def trigger_args(self, args, tail=None, allowed=None, forward=None):
        command = args.pop(0).upper()

        if allowed is not None and command not in allowed:
            raise CommandParserError(f"Illegal command supplied: '{command}'")

        if command in self._commands:
            (cmd, func) = self._commands[command]
            cmd_args = cmd.parse_args(args)
            cmd_args._tail = tail
            cmd_args._forward = forward
            await func(cmd_args)
        elif command == "HELP":
            out = ["Following commands are supported:", ""]
            for name, (cmd, func) in self._commands.items():
                if cmd.prog == name:
                    out.append("\t{} - {}".format(cmd.prog, cmd.short_description))

            out.append("")
            out.append("To get more help, add -h to any command without arguments.")

            raise CommandParserError("\n".join(out))
        else:
            raise CommandParserError('Unknown command "{}", type HELP for list'.format(command))

    async def trigger(self, text, tail=None, allowed=None, forward=None):
        for args in split(text):
            await self.trigger_args(args, tail, allowed, forward)
            tail = None



import asyncio
import logging

"""
Buffering event queue with merging of events.
"""


class EventQueue:
    def __init__(self, callback):
        self._callback = callback
        self._events = []
        self._loop = asyncio.get_running_loop()
        self._timer = None
        self._start = 0
        self._chain = asyncio.Queue()
        self._task = None
        self._timeout = 3600

    def start(self):
        if self._task is None:
            self._task = asyncio.ensure_future(self._run())

    def stop(self):
        if self._task:
            self._task.cancel()
            self._task = None

    async def _run(self):
        while True:
            try:
                task = await self._chain.get()
            except asyncio.CancelledError:
                logging.debug("EventQueue was cancelled.")
                return

            try:
                await asyncio.wait_for(task, timeout=self._timeout)
            except asyncio.CancelledError:
                logging.debug("EventQueue task was cancelled.")
                return
            except asyncio.TimeoutError:
                logging.warning("EventQueue task timed out.")
            finally:
                self._chain.task_done()

    def _flush(self):
        events = self._events

        self._timer = None
        self._events = []

        self._chain.put_nowait(self._callback(events))

    def enqueue(self, event):
        now = self._loop.time()

        # always cancel timer when we enqueue
        if self._timer:
            self._timer.cancel()

        # stamp start time when we queue first event, always append event
        if len(self._events) == 0:
            self._start = now
            self._events.append(event)
        else:
            # lets see if we can merge the event
            prev = self._events[-1]

            prev_formatted = "format" in prev["content"]
            cur_formatted = "format" in event["content"]

            # calculate content length if we need to flush anyway to stay within max event size
            prev_len = 0
            if "content" in prev:
                if "body" in prev["content"]:
                    prev_len += len(prev["content"]["body"])
                if "formatted_body" in prev["content"]:
                    prev_len += len(prev["content"]["formatted_body"])

            if (
                prev["type"] == event["type"]
                and prev["type"][0] != "_"
                and prev["user_id"] == event["user_id"]
                and "msgtype" in prev["content"]
                and prev["content"]["msgtype"] == event["content"]["msgtype"]
                and prev_formatted == cur_formatted
                and prev_len < 64_000  # a single IRC event can't overflow with this
            ):
                prev["content"]["body"] += "\n" + event["content"]["body"]
                if cur_formatted:
                    prev["content"]["formatted_body"] += "<br>" + event["content"]["formatted_body"]
            else:
                # can't merge, force flush but enqueue the next event
                self._flush()
                self._start = now
                self._events.append(event)

        # if we have bumped ourself for a full second, flush now
        if now >= self._start + 1.0:
            self._flush()
        else:
            self._timer = self._loop.call_later(0.1, self._flush)



import asyncio
import re
from argparse import Namespace
from html import escape
from urllib.parse import urlparse

from mautrix.errors import MatrixRequestError

from heisenbridge import __version__
from heisenbridge.command_parse import CommandManager
from heisenbridge.command_parse import CommandParser
from heisenbridge.command_parse import CommandParserError
from heisenbridge.network_room import NetworkRoom
from heisenbridge.room import Room
from heisenbridge.room import RoomInvalidError


def indent(n):
    return "&nbsp;" * n * 8


class ControlRoom(Room):
    commands: CommandManager

    def init(self):
        self.commands = CommandManager()

        cmd = CommandParser(prog="NETWORKS", description="list available networks")
        self.commands.register(cmd, self.cmd_networks)

        cmd = CommandParser(prog="SERVERS", description="list servers for a network")
        cmd.add_argument("network", help="network name (see NETWORKS)")
        self.commands.register(cmd, self.cmd_servers)

        cmd = CommandParser(prog="OPEN", description="open network for connecting")
        cmd.add_argument("name", help="network name (see NETWORKS)")
        cmd.add_argument("--new", action="store_true", help="force open a new network connection")
        self.commands.register(cmd, self.cmd_open)

        cmd = CommandParser(
            prog="STATUS",
            description="show bridge status",
            epilog="Note: admins see all users but only their own rooms",
        )
        self.commands.register(cmd, self.cmd_status)

        cmd = CommandParser(
            prog="QUIT",
            description="disconnect from all networks",
            epilog=(
                "For quickly leaving all networks and removing configurations in a single command.\n"
                "\n"
                "Additionally this will close current DM session with the bridge.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_quit)

        if self.serv.is_admin(self.user_id):
            cmd = CommandParser(prog="MASKS", description="list allow masks")
            self.commands.register(cmd, self.cmd_masks)

            cmd = CommandParser(
                prog="HIDDENROOM",
                description="Use a hidden room to offload invites into. Keeps room history clean.",
            )
            group = cmd.add_mutually_exclusive_group()
            group.add_argument("--enable", help="Enable use of hidden room", action="store_true")
            group.add_argument("--disable", help="Disable use of hidden room", action="store_true")
            self.commands.register(cmd, self.cmd_hidden_room)

            cmd = CommandParser(
                prog="ADDMASK",
                description="add new allow mask",
                epilog=(
                    "For anyone else than the owner to use this bridge they need to be allowed to talk with the bridge bot.\n"
                    "This is accomplished by adding an allow mask that determines their permission level when using the bridge.\n"
                    "\n"
                    "Only admins can manage networks, normal users can just connect.\n"
                ),
            )
            cmd.add_argument("mask", help="Matrix ID mask (eg: @friend:contoso.com or *:contoso.com)")
            cmd.add_argument("--admin", help="Admin level access", action="store_true")
            self.commands.register(cmd, self.cmd_addmask)

            cmd = CommandParser(
                prog="DELMASK",
                description="delete allow mask",
                epilog=(
                    "Note: Removing a mask only prevents starting a new DM with the bridge bot. Use FORGET for ending existing"
                    " sessions."
                ),
            )
            cmd.add_argument("mask", help="Matrix ID mask (eg: @friend:contoso.com or *:contoso.com)")
            self.commands.register(cmd, self.cmd_delmask)

            cmd = CommandParser(prog="ADDNETWORK", description="add new network")
            cmd.add_argument("name", help="network name")
            self.commands.register(cmd, self.cmd_addnetwork)

            cmd = CommandParser(prog="DELNETWORK", description="delete network")
            cmd.add_argument("name", help="network name")
            self.commands.register(cmd, self.cmd_delnetwork)

            cmd = CommandParser(prog="ADDSERVER", description="add server to a network")
            cmd.add_argument("network", help="network name")
            cmd.add_argument("address", help="server address")
            cmd.add_argument("port", nargs="?", type=int, help="server port", default=6667)
            cmd.add_argument("--tls", action="store_true", help="use TLS encryption", default=False)
            cmd.add_argument(
                "--tls-insecure",
                action="store_true",
                help="ignore TLS verification errors (hostname, self-signed, expired)",
                default=False,
            )
            cmd.add_argument(
                "--tls-ciphers",
                help="set TLS cipher string (in OpenSSL cipher list format)",
                default=None,
            )
            cmd.add_argument("--proxy", help="use a SOCKS proxy (socks5://...)", default=None)
            self.commands.register(cmd, self.cmd_addserver)

            cmd = CommandParser(prog="DELSERVER", description="delete server from a network")
            cmd.add_argument("network", help="network name")
            cmd.add_argument("address", help="server address")
            cmd.add_argument("port", nargs="?", type=int, help="server port", default=6667)
            self.commands.register(cmd, self.cmd_delserver)

            cmd = CommandParser(
                prog="FORGET",
                description="remove all connections and configuration of a user",
                epilog=(
                    "Kills all connections of this user, removes all user set configuration and makes the bridge leave all rooms"
                    " where this user is in.\n"
                    "If the user still has an allow mask they can DM the bridge again to reconfigure and reconnect.\n"
                    "\n"
                    "This is meant as a way to kick users after removing an allow mask or resetting a user after losing access to"
                    " existing account/rooms for any reason.\n"
                ),
            )
            cmd.add_argument("user", help="Matrix ID (eg: @ex-friend:contoso.com)")
            self.commands.register(cmd, self.cmd_forget)

            cmd = CommandParser(prog="DISPLAYNAME", description="change bridge displayname")
            cmd.add_argument("displayname", help="new bridge displayname")
            self.commands.register(cmd, self.cmd_displayname)

            cmd = CommandParser(prog="AVATAR", description="change bridge avatar")
            cmd.add_argument("url", help="new avatar URL (mxc:// format)")
            self.commands.register(cmd, self.cmd_avatar)

            cmd = CommandParser(
                prog="IDENT",
                description="configure ident replies",
                epilog="Note: MXID here is case sensitive, see subcommand help with IDENTCFG SET -h",
            )
            subcmd = cmd.add_subparsers(help="commands", dest="cmd")
            subcmd.add_parser("list", help="list custom idents (default)")
            cmd_set = subcmd.add_parser("set", help="set custom ident")
            cmd_set.add_argument("mxid", help="mxid of the user")
            cmd_set.add_argument("ident", help="custom ident for the user")
            cmd_remove = subcmd.add_parser("remove", help="remove custom ident")
            cmd_remove.add_argument("mxid", help="mxid of the user")
            self.commands.register(cmd, self.cmd_ident)

            cmd = CommandParser(
                prog="SYNC",
                description="set default IRC member sync mode",
                epilog="Note: Users can override this per room.",
            )
            group = cmd.add_mutually_exclusive_group()
            group.add_argument("--lazy", help="set lazy sync, members are added when they talk", action="store_true")
            group.add_argument(
                "--half", help="set half sync, members are added when they join or talk (default)", action="store_true"
            )
            group.add_argument("--full", help="set full sync, members are fully synchronized", action="store_true")
            self.commands.register(cmd, self.cmd_sync)

            cmd = CommandParser(
                prog="MAXLINES",
                description="set default maximum number of lines per message until truncation or pastebin",
                epilog="Note: Users can override this per room.",
            )
            cmd.add_argument("lines", type=int, nargs="?", help="Number of lines")
            self.commands.register(cmd, self.cmd_maxlines)

            cmd = CommandParser(
                prog="PASTEBIN",
                description="enable or disable automatic pastebin of long messages by default",
                epilog="Note: Users can override this per room.",
            )
            cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable pastebin")
            cmd.add_argument(
                "--disable", dest="enabled", action="store_false", help="Disable pastebin (messages will be truncated)"
            )
            cmd.set_defaults(enabled=None)
            self.commands.register(cmd, self.cmd_pastebin)

            cmd = CommandParser(
                prog="REACTS",
                description="enable or disable reacting to messages on splits/linking",
                epilog="Note: Users can override this per room.",
            )
            cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable reacts")
            cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable reacts")
            cmd.set_defaults(enabled=None)
            self.commands.register(cmd, self.cmd_reacts)

            cmd = CommandParser(prog="MEDIAURL", description="configure media URL for links")
            cmd.add_argument("url", nargs="?", help="new URL override")
            cmd.add_argument("--remove", help="remove URL override (will retry auto-detection)", action="store_true")
            self.commands.register(cmd, self.cmd_media_url)

            cmd = CommandParser(prog="MEDIAPATH", description="configure media path for links")
            cmd.add_argument("path", nargs="?", help="new path override")
            cmd.add_argument("--remove", help="remove path override", action="store_true")
            self.commands.register(cmd, self.cmd_media_path)

            cmd = CommandParser(prog="VERSION", description="show bridge version")
            self.commands.register(cmd, self.cmd_version)

        self.mx_register("m.room.message", self.on_mx_message)

    def is_valid(self) -> bool:
        if self.user_id is None:
            return False

        if len(self.members) != 2:
            return False

        return True

    async def show_help(self):
        self.send_notice_html(
            f"<b>Howdy, stranger!</b> You have been granted access to the IRC bridge of <b>{self.serv.server_name}</b>."
        )

        try:
            return await self.commands.trigger("HELP")
        except CommandParserError as e:
            return self.send_notice(str(e))

    async def on_mx_message(self, event) -> bool:
        if str(event.content.msgtype) != "m.text" or event.sender == self.serv.user_id:
            return

        # ignore edits
        if event.content.get_edit():
            return

        try:
            if event.content.formatted_body:
                lines = str(await self.parser.parse(event.content.formatted_body)).split("\n")
            else:
                lines = event.content.body.split("\n")

            command = lines.pop(0)
            tail = "\n".join(lines) if len(lines) > 0 else None

            await self.commands.trigger(command, tail)
        except CommandParserError as e:
            self.send_notice(str(e))

    def networks(self):
        networks = {}

        for network, config in self.serv.config["networks"].items():
            config["name"] = network
            networks[network.lower()] = config

        return networks

    async def cmd_hidden_room(self, args):
        if args.enable:
            self.serv.config["use_hidden_room"] = True
            await self.serv.save()
        elif args.disable:
            self.serv.config["use_hidden_room"] = False
            await self.serv.save()

        try:
            is_enabled = await self.serv.ensure_hidden_room()
            self.send_notice(f"Hidden room is {'enabled' if is_enabled else 'disabled'}.")
        except Exception as e:
            self.send_notice(f"Failed setting up hidden room: {e}")

    async def cmd_masks(self, args):
        msg = "Configured masks:\n"

        for mask, value in self.serv.config["allow"].items():
            msg += "\t{} -> {}\n".format(mask, value)

        self.send_notice(msg)

    async def cmd_addmask(self, args):
        masks = self.serv.config["allow"]

        if args.mask in masks:
            return self.send_notice("Mask already exists")

        masks[args.mask] = "admin" if args.admin else "user"
        await self.serv.save()

        self.send_notice("Mask added.")

    async def cmd_delmask(self, args):
        masks = self.serv.config["allow"]

        if args.mask not in masks:
            return self.send_notice("Mask does not exist")

        del masks[args.mask]
        await self.serv.save()

        self.send_notice("Mask removed.")

    async def cmd_networks(self, args):
        networks = self.serv.config["networks"]

        self.send_notice("Configured networks:")

        for network, data in networks.items():
            self.send_notice(f"\t{network} ({len(data['servers'])} servers)")

    async def cmd_addnetwork(self, args):
        networks = self.networks()

        if args.name.lower() in networks:
            return self.send_notice("Network already exists")

        self.serv.config["networks"][args.name] = {"servers": []}
        await self.serv.save()

        self.send_notice("Network added.")

    async def cmd_delnetwork(self, args):
        networks = self.networks()

        if args.name.lower() not in networks:
            return self.send_notice("Network does not exist")

        # FIXME: check if anyone is currently connected

        # FIXME: if no one is currently connected, leave from all network related rooms

        del self.serv.config["networks"][args.name]
        await self.serv.save()

        return self.send_notice("Network removed.")

    async def cmd_servers(self, args):
        networks = self.networks()

        if args.network.lower() not in networks:
            return self.send_notice("Network does not exist")

        network = networks[args.network.lower()]

        self.send_notice(f"Configured servers for {network['name']}:")

        for server in network["servers"]:
            with_tls = ""
            if server["tls"]:
                if "tls_insecure" in server and server["tls_insecure"]:
                    with_tls = "with insecure TLS"
                else:
                    with_tls = "with TLS"
            proxy = (
                f" through {server['proxy']}"
                if "proxy" in server and server["proxy"] is not None and len(server["proxy"]) > 0
                else ""
            )
            self.send_notice(f"\t{server['address']}:{server['port']} {with_tls}{proxy}")

    async def cmd_addserver(self, args):
        networks = self.networks()

        if args.network.lower() not in networks:
            return self.send_notice("Network does not exist")

        network = networks[args.network.lower()]
        address = args.address.lower()

        for server in network["servers"]:
            if server["address"] == address and server["port"] == args.port:
                return self.send_notice("This server already exists.")

        self.serv.config["networks"][network["name"]]["servers"].append(
            {
                "address": address,
                "port": args.port,
                "tls": args.tls,
                "tls_insecure": args.tls_insecure,
                "tls_ciphers": args.tls_ciphers,
                "proxy": args.proxy,
            }
        )
        await self.serv.save()

        self.send_notice("Server added.")

    async def cmd_delserver(self, args):
        networks = self.networks()

        if args.network.lower() not in networks:
            return self.send_notice("Network does not exist")

        network = networks[args.network.lower()]
        address = args.address.lower()

        to_pop = -1
        for i, server in enumerate(network["servers"]):
            if server["address"] == address and server["port"] == args.port:
                to_pop = i
                break

        if to_pop == -1:
            return self.send_notice("No such server.")

        self.serv.config["networks"][network["name"]]["servers"].pop(to_pop)
        await self.serv.save()

        self.send_notice("Server deleted.")

    async def cmd_status(self, args):
        users = set()

        if self.serv.is_admin(self.user_id):
            for room in self.serv.find_rooms():
                if room.user_id is not None:  # ignore HiddenRoom
                    users.add(room.user_id)

            users = list(users)
            users.sort()
        else:
            users.add(self.user_id)

        self.send_notice_html(f"I have {len(users)} known users:")
        for user_id in users:
            ncontrol = len(self.serv.find_rooms("ControlRoom", user_id))

            self.send_notice_html(f"{indent(1)}{user_id} ({ncontrol} open control rooms):")

            for network in self.serv.find_rooms("NetworkRoom", user_id):
                connected = "not connected"
                channels = "not in channels"
                privates = "not in PMs"
                plumbs = "not in plumbs"

                if network.conn and network.conn.connected:
                    user = network.real_user if network.real_user[0] != "?" else "?"
                    host = network.real_host if network.real_host[0] != "?" else "?"
                    connected = f"connected as {network.conn.real_nickname}!{user}@{host}"

                nchannels = 0
                nprivates = 0
                nplumbs = 0

                for room in network.rooms.values():
                    if type(room).__name__ == "PrivateRoom":
                        nprivates += 1
                    if type(room).__name__ == "ChannelRoom":
                        nchannels += 1
                    if type(room).__name__ == "PlumbedRoom":
                        nplumbs += 1

                if nprivates > 0:
                    privates = f"in {nprivates} PMs"

                if nchannels > 0:
                    channels = f"in {nchannels} channels"

                if nplumbs > 0:
                    plumbs = f"in {nplumbs} plumbs"

                self.send_notice_html(f"{indent(2)}{network.name}, {connected}, {channels}, {privates}, {plumbs}")

                if self.user_id == user_id:
                    for room in network.rooms.values():
                        join = ""
                        if not room.in_room(user_id):
                            join = " (you have not joined this room)"
                            # ensure the user invite is valid
                            await self.az.intent.invite_user(room.id, self.user_id)
                        self.send_notice_html(
                            f'{indent(3)}<a href="https://matrix.to/#/{escape(room.id)}">{escape(room.name)}</a>{join}'
                        )

    async def cmd_forget(self, args):
        if args.user == self.user_id:
            return self.send_notice("I can't forget you, silly!")

        rooms = self.serv.find_rooms(None, args.user)

        if len(rooms) == 0:
            return self.send_notice("No such user. See STATUS for list of users.")

        # disconnect each network room in first pass
        for room in rooms:
            if type(room) == NetworkRoom and room.conn and room.conn.connected:
                self.send_notice(f"Disconnecting {args.user} from {room.name}...")
                await room.cmd_disconnect(Namespace())

        self.send_notice(f"Leaving all {len(rooms)} rooms {args.user} was in...")

        # then just forget everything
        for room in rooms:
            self.serv.unregister_room(room.id)

            try:
                await self.az.intent.leave_room(room.id)
            except MatrixRequestError:
                pass
            try:
                await self.az.intent.forget_room(room.id)
            except MatrixRequestError:
                pass

        self.send_notice(f"Done, I have forgotten about {args.user}")

    async def cmd_displayname(self, args):
        try:
            await self.az.intent.set_displayname(args.displayname)
        except MatrixRequestError as e:
            self.send_notice(f"Failed to set displayname: {str(e)}")

    async def cmd_avatar(self, args):
        try:
            await self.az.intent.set_avatar_url(args.url)
        except MatrixRequestError as e:
            self.send_notice(f"Failed to set avatar: {str(e)}")

    async def cmd_ident(self, args):
        idents = self.serv.config["idents"]

        if args.cmd == "list" or args.cmd is None:
            self.send_notice("Configured custom idents:")
            for mxid, ident in idents.items():
                self.send_notice(f"\t{mxid} -> {ident}")
        elif args.cmd == "set":
            if not re.match(r"^[a-z][-a-z0-9]*$", args.ident):
                self.send_notice(f"Invalid ident string: {args.ident}")
                self.send_notice("Must be lowercase, start with a letter, can contain dashes, letters and numbers.")
            else:
                idents[args.mxid] = args.ident
                self.send_notice(f"Set custom ident for {args.mxid} to {args.ident}")
                await self.serv.save()
        elif args.cmd == "remove":
            if args.mxid in idents:
                del idents[args.mxid]
                self.send_notice(f"Removed custom ident for {args.mxid}")
                await self.serv.save()
            else:
                self.send_notice(f"No custom ident for {args.mxid}")

    async def cmd_sync(self, args):
        if args.lazy:
            self.serv.config["member_sync"] = "lazy"
            await self.serv.save()
        elif args.half:
            self.serv.config["member_sync"] = "half"
            await self.serv.save()
        elif args.full:
            self.serv.config["member_sync"] = "full"
            await self.serv.save()

        self.send_notice(f"Member sync is set to {self.serv.config['member_sync']}")

    async def cmd_media_url(self, args):
        if args.remove:
            self.serv.config["media_url"] = None
            await self.serv.save()
            self.serv.endpoint = await self.serv.detect_public_endpoint()
        elif args.url:
            parsed = urlparse(args.url)
            if parsed.scheme in ["http", "https"] and not parsed.params and not parsed.query and not parsed.fragment:
                self.serv.config["media_url"] = args.url
                await self.serv.save()
                self.serv.endpoint = args.url
            else:
                self.send_notice(f"Invalid media URL format: {args.url}")
                return

        self.send_notice(f"Media URL override is set to {self.serv.config['media_url']}")
        self.send_notice(f"Current active media URL: {self.serv.endpoint}")

    async def cmd_media_path(self, args):
        if args.remove:
            self.serv.config["media_path"] = None
            await self.serv.save()
            self.serv.media_path = self.serv.DEFAULT_MEDIA_PATH
        elif args.path:
            self.serv.config["media_path"] = args.path
            await self.serv.save()
            self.serv.media_path = args.path

        self.send_notice(f"Media Path override is set to {self.serv.config['media_path']}")
        self.send_notice(f"Current active media path: {self.serv.media_path}")

    async def cmd_maxlines(self, args):
        if args.lines is not None:
            self.serv.config["max_lines"] = args.lines
            await self.serv.save()

        self.send_notice(f"Max lines default is {self.serv.config['max_lines']}")

    async def cmd_pastebin(self, args):
        if args.enabled is not None:
            self.serv.config["use_pastebin"] = args.enabled
            await self.serv.save()

        self.send_notice(f"Pastebin is {'enabled' if self.serv.config['use_pastebin'] else 'disabled'} by default")

    async def cmd_reacts(self, args):
        if args.enabled is not None:
            self.serv.config["use_reacts"] = args.enabled
            await self.serv.save()

        self.send_notice(f"Reacts are {'enabled' if self.serv.config['use_reacts'] else 'disabled'} by default")

    async def cmd_open(self, args):
        networks = self.networks()
        name = args.name.lower()

        if name not in networks:
            return self.send_notice("Network does not exist")

        network = networks[name]

        found = 0
        for room in self.serv.find_rooms(NetworkRoom, self.user_id):
            if room.name == network["name"]:
                found += 1

                if not args.new:
                    if self.user_id not in room.members:
                        self.send_notice(f"Inviting back to {room.name} ({room.id})")
                        await self.az.intent.invite_user(room.id, self.user_id)
                    else:
                        self.send_notice(f"You are already in {room.name} ({room.id})")

        # if we found at least one network room, no need to create unless forced
        if found > 0 and not args.new:
            return

        name = network["name"] if found == 0 else f"{network['name']} {found + 1}"

        self.send_notice(f"You have been invited to {name}")
        await NetworkRoom.create(self.serv, network["name"], self.user_id, name)

    async def cmd_quit(self, args):
        rooms = self.serv.find_rooms(None, self.user_id)

        # disconnect each network room in first pass
        for room in rooms:
            if type(room) == NetworkRoom and room.conn and room.conn.connected:
                self.send_notice(f"Disconnecting from {room.name}...")
                await room.cmd_disconnect(Namespace())

        self.send_notice("Closing all channels and private messages...")

        # then just forget everything
        for room in rooms:
            if room.id == self.id:
                continue

            self.serv.unregister_room(room.id)

            try:
                await self.az.intent.leave_room(room.id)
            except MatrixRequestError:
                pass
            try:
                await self.az.intent.forget_room(room.id)
            except MatrixRequestError:
                pass

        self.send_notice("Goodbye!")
        await asyncio.sleep(1)
        raise RoomInvalidError("Leaving")

    async def cmd_version(self, args):
        self.send_notice(f"heisenbridge v{__version__}")



import asyncio
import logging
import re
from typing import Optional

from irc.modes import parse_channel_modes
from mautrix.errors import MatrixRequestError
from mautrix.types import Membership
from mautrix.types import MessageEvent
from mautrix.types import MessageType
from mautrix.types import TextMessageEventContent

from heisenbridge.channel_room import ChannelRoom
from heisenbridge.command_parse import CommandParser
from heisenbridge.private_room import parse_irc_formatting


class NetworkRoom:
    pass


def connected(f):
    def wrapper(*args, **kwargs):
        self = args[0]

        if not self.network or not self.network.conn or not self.network.conn.connected:
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


def send_relaymsg(room, func, sender):
    def wrapper(target, message):
        message = f":\001ACTION {message}\001" if func == room.network.conn.action else f":{message}"
        room.network.conn.send_items("RELAYMSG", target, f"{sanitize_irc_nick(sender)}/{room.relaytag}", message)

    return wrapper


def sanitize_irc_nick(nick):
    return re.sub(r"[^A-Za-z0-9_\[\]|{}-]", "", nick)


class PlumbedRoom(ChannelRoom):
    max_lines = 5
    use_pastebin = True
    use_reacts = True
    use_displaynames = True
    use_disambiguation = True
    use_zwsp = False
    allow_notice = False
    force_forward = True
    topic_sync = None
    relaytag = "m"

    def init(self) -> None:
        super().init()

        cmd = CommandParser(
            prog="DISPLAYNAMES", description="enable or disable use of displaynames in relayed messages"
        )
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable displaynames")
        cmd.add_argument(
            "--disable", dest="enabled", action="store_false", help="Disable displaynames (fallback to MXID)"
        )
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_displaynames)

        cmd = CommandParser(
            prog="DISAMBIGUATION", description="enable or disable disambiguation of conflicting displaynames"
        )
        cmd.add_argument(
            "--enable", dest="enabled", action="store_true", help="Enable disambiguation (postfix with MXID)"
        )
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable disambiguation")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_disambiguation)

        cmd = CommandParser(prog="ZWSP", description="enable or disable Zero-Width-Space anti-ping")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable ZWSP anti-ping")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable ZWSP anti-ping")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_zwsp)

        cmd = CommandParser(prog="NOTICERELAY", description="enable or disable relaying of Matrix notices to IRC")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable notice relay")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable notice relay")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_noticerelay)

        cmd = CommandParser(prog="TOPIC", description="show or set channel topic and configure sync mode")
        cmd.add_argument("--sync", choices=["off", "irc", "matrix", "any"], help="Topic sync targets, defaults to off")
        cmd.add_argument("text", nargs="*", help="topic text if setting")
        self.commands.register(cmd, self.cmd_topic)

        cmd = CommandParser(prog="RELAYTAG", description="set RELAYMSG tag if supported by server")
        cmd.add_argument("tag", nargs="?", help="new tag")
        self.commands.register(cmd, self.cmd_relaytag)

        self.mx_register("m.room.topic", self._on_mx_room_topic)

    def is_valid(self) -> bool:
        # we are valid as long as the appservice is in the room
        if not self.in_room(self.serv.user_id):
            return False

        return True

    @staticmethod
    async def create(network: "NetworkRoom", id: str, channel: str, key: str) -> "ChannelRoom":
        logging.debug(f"PlumbedRoom.create(network='{network.name}', id='{id}', channel='{channel}', key='{key}'")

        network.send_notice(f"Joining room {id} to initiate plumb...")
        try:
            room_id = await network.az.intent.join_room(id)
        except MatrixRequestError as e:
            network.send_notice(f"Failed to join room: {str(e)}")
            return

        network.send_notice(f"Joined room {room_id}, refreshing member state...")
        await network.az.intent.get_room_members(room_id)
        network.send_notice(f"Got state for room {room_id}, plumbing...")

        joined = await network.az.state_store.get_member_profiles(room_id, (Membership.JOIN,))
        banned = await network.az.state_store.get_members(room_id, (Membership.BAN,))

        room = PlumbedRoom(room_id, network.user_id, network.serv, joined, banned)
        room.name = channel.lower()
        room.key = key
        room.network = network
        room.network_id = network.id
        room.network_name = network.name

        # stamp global member sync setting at room creation time
        room.member_sync = network.serv.config["member_sync"]

        room.max_lines = network.serv.config["max_lines"]
        room.use_pastebin = network.serv.config["use_pastebin"]
        room.use_reacts = network.serv.config["use_reacts"]

        for user_id, member in joined.items():
            if member.displayname is not None:
                room.displaynames[user_id] = member.displayname

        network.serv.register_room(room)
        network.rooms[room.name] = room
        await room.save()

        network.send_notice(f"Plumbed {room_id} to {channel}, to unplumb just kick me out.")
        return room

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "use_displaynames" in config:
            self.use_displaynames = config["use_displaynames"]

        if "use_disambiguation" in config:
            self.use_disambiguation = config["use_disambiguation"]

        if "use_zwsp" in config:
            self.use_zwsp = config["use_zwsp"]

        if "allow_notice" in config:
            self.allow_notice = config["allow_notice"]

        if "topic_sync" in config:
            self.topic_sync = config["topic_sync"]

        if "relaytag" in config:
            self.relaytag = config["relaytag"]

    def to_config(self) -> dict:
        return {
            **(super().to_config()),
            "use_displaynames": self.use_displaynames,
            "use_disambiguation": self.use_disambiguation,
            "use_zwsp": self.use_zwsp,
            "allow_notice": self.allow_notice,
            "topic_sync": self.topic_sync,
            "relaytag": self.relaytag,
        }

    # topic updates from channel state replies are ignored because formatting changes
    def set_topic(self, topic: str, user_id: Optional[str] = None) -> None:
        pass

    def on_topic(self, conn, event) -> None:
        self.send_notice("{} changed the topic".format(event.source.nick))
        if conn.real_nickname != event.source.nick and self.topic_sync in ["matrix", "any"]:
            (plain, formatted) = parse_irc_formatting(event.arguments[0])
            super().set_topic(plain)

    @connected
    async def _on_mx_room_topic(self, event) -> None:
        if event.sender != self.serv.user_id and self.topic_sync in ["irc", "any"]:
            topic = re.sub(r"[\r\n]", " ", event.content.topic)
            self.network.conn.topic(self.name, topic)

    @connected
    async def on_mx_message(self, event) -> None:
        sender = str(event.sender)
        (name, server) = sender.split(":", 1)

        # ignore self messages
        if sender == self.serv.user_id:
            return

        # prevent re-sending federated messages back
        if name.startswith("@" + self.serv.puppet_prefix) and server == self.serv.server_name:
            return

        # add ZWSP to sender to avoid pinging on IRC
        if self.use_zwsp:
            sender = f"{name[:2]}\u200B{name[2:]}:{server[:1]}\u200B{server[1:]}"

        if self.use_displaynames and event.sender in self.displaynames:
            sender_displayname = self.displaynames[event.sender]

            # ensure displayname is unique
            if self.use_disambiguation:
                for user_id, displayname in self.displaynames.items():
                    if user_id != event.sender and displayname == sender_displayname:
                        sender_displayname += f" ({sender})"
                        break

            # add ZWSP if displayname matches something on IRC
            if self.use_zwsp and len(sender_displayname) > 1:
                sender_displayname = f"{sender_displayname[:1]}\u200B{sender_displayname[1:]}"

            sender = sender_displayname

        # limit plumbed sender max length to 100 characters
        sender = sender[:100]

        if event.content.msgtype.is_media:
            # process media event like it was a text message
            media_event = MessageEvent(
                sender=event.sender,
                type=None,
                room_id=None,
                event_id=None,
                timestamp=None,
                content=TextMessageEventContent(body=self.serv.mxc_to_url(event.content.url, event.content.body)),
            )
            await self.relay_message(media_event, self.network.conn.privmsg, sender)

            if self.use_reacts:
                self.react(event.event_id, "\U0001F517")  # link
            self.media.append([event.event_id, event.content.url])
            await self.save()
        elif event.content.msgtype == MessageType.EMOTE:
            await self.relay_message(event, self.network.conn.action, sender)
        elif event.content.msgtype == MessageType.TEXT:
            await self.relay_message(event, self.network.conn.privmsg, sender)
        elif event.content.msgtype == MessageType.NOTICE and self.allow_notice:
            await self.relay_message(event, self.network.conn.notice, sender)

        await self.az.intent.send_receipt(event.room_id, event.event_id)

    async def relay_message(self, event, func, sender):
        prefix = f"{sender} " if str(event.content.msgtype) == "m.emote" else f"<{sender}> "

        # if we have relaymsg cap and it's not a notice, add more abstraction
        if "draft/relaymsg" in self.network.caps_enabled and event.content.msgtype != MessageType.NOTICE:
            func = send_relaymsg(self, func, sender)
            prefix = None

        await self._send_message(event, func, prefix)

    @connected
    async def on_mx_ban(self, user_id) -> None:
        nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
        if nick is None:
            return

        # best effort kick and ban
        self.network.conn.mode(self.name, f"+b {nick}!*@*")
        self.network.conn.kick(self.name, nick, "You have been banned on Matrix")

    @connected
    async def on_mx_unban(self, user_id) -> None:
        nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
        if nick is None:
            return

        # best effort unban
        self.network.conn.mode(self.name, f"-b {nick}!*@*")

    @connected
    async def on_mx_leave(self, user_id) -> None:
        nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
        if nick is None:
            return

        # best effort kick
        if self.is_on_channel(nick):
            self.network.conn.kick(self.name, nick, "You have been kicked on Matrix")

    def pills(self):
        # if pills are disabled, don't generate any
        if self.network.pills_length < 1:
            return None

        ret = super().pills()

        # remove the bot from pills as it may cause confusion
        nick = self.network.conn.real_nickname.lower()
        if nick in ret:
            del ret[nick]

        return ret

    async def cmd_displaynames(self, args) -> None:
        if args.enabled is not None:
            self.use_displaynames = args.enabled
            await self.save()

        self.send_notice(f"Displaynames are {'enabled' if self.use_displaynames else 'disabled'}")

    async def cmd_disambiguation(self, args) -> None:
        if args.enabled is not None:
            self.use_disambiguation = args.enabled
            await self.save()

        self.send_notice(f"Dismabiguation is {'enabled' if self.use_disambiguation else 'disabled'}")

    async def cmd_zwsp(self, args) -> None:
        if args.enabled is not None:
            self.use_zwsp = args.enabled
            await self.save()

        self.send_notice(f"Zero-Width-Space anti-ping is {'enabled' if self.use_zwsp else 'disabled'}")

    async def cmd_noticerelay(self, args) -> None:
        if args.enabled is not None:
            self.allow_notice = args.enabled
            await self.save()

        self.send_notice(f"Notice relay is {'enabled' if self.allow_notice else 'disabled'}")

    async def cmd_topic(self, args) -> None:
        if args.sync is None:
            self.network.conn.topic(self.name, " ".join(args.text))
            return

        self.topic_sync = args.sync if args.sync != "off" else None
        self.send_notice(f"Topic sync is {self.topic_sync if self.topic_sync else 'off'}")
        await self.save()

    async def cmd_relaytag(self, args) -> None:
        if args.tag:
            self.relaytag = sanitize_irc_nick(args.tag)
            await self.save()

        self.send_notice(f"Relay tag is '{self.relaytag}'")

    def on_mode(self, conn, event) -> None:
        super().on_mode(conn, event)

        # when we get ops (or half-ops) get current ban list to see if we need to ban someone that has been banned on matrix
        modes = list(event.arguments)
        for sign, key, value in parse_channel_modes(" ".join(modes)):
            if sign == "+" and key in ["o", "h"] and value == self.network.conn.real_nickname:
                self.network.conn.mode(self.name, "+b")

    def on_endofbanlist(self, conn, event) -> None:
        masks = [ban[0].lower() for ban in self.bans_buffer]
        super().on_endofbanlist(conn, event)

        # add any nick bans that are missing from IRC
        for user_id in self.bans:
            nick = self.serv.nick_from_irc_user_id(self.network.name, user_id)
            if nick is None:
                continue

            mask = f"{nick}!*@*"
            if mask not in masks:
                self.network.conn.mode(self.name, f"+b {mask}")
            if self.is_on_channel(nick):
                self.network.conn.kick(self.name, nick)



import asyncio
import collections
import html
import logging
import re
import unicodedata
from datetime import datetime
from datetime import timezone
from html import escape
from typing import List
from typing import Optional
from typing import Tuple
from urllib.parse import urlparse

from mautrix.api import Method
from mautrix.api import SynapseAdminPath
from mautrix.errors import MatrixStandardRequestError
from mautrix.types.event.state import JoinRestriction
from mautrix.types.event.state import JoinRestrictionType
from mautrix.types.event.state import JoinRule
from mautrix.types.event.state import JoinRulesStateEventContent
from mautrix.types.event.type import EventType

from heisenbridge.command_parse import CommandManager
from heisenbridge.command_parse import CommandParser
from heisenbridge.command_parse import CommandParserError
from heisenbridge.room import Room


class NetworkRoom:
    pass


def unix_to_local(timestamp: Optional[str]):
    try:
        dt = datetime.fromtimestamp(int(timestamp), timezone.utc)
        return dt.strftime("%c %Z")  # intentionally UTC for now
    except ValueError:
        logging.debug("Tried to convert '{timestamp}' to int")
        return timestamp


def connected(f):
    def wrapper(*args, **kwargs):
        self = args[0]

        if not self.network or not self.network.conn or not self.network.conn.connected:
            self.send_notice("Need to be connected to use this command.")
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


def parse_irc_formatting(input: str, pills=None, color=None) -> Tuple[str, Optional[str]]:
    plain = []
    formatted = []

    color_table = collections.defaultdict(
        lambda: None,
        {
            "0": "#ffffff",
            "00": "#ffffff",
            "1": "#000000",
            "01": "#000000",
            "2": "#00007f",
            "02": "#00007f",
            "3": "#009300",
            "03": "#009300",
            "4": "#ff0000",
            "04": "#ff0000",
            "5": "#7f0000",
            "05": "#7f0000",
            "6": "#9c009c",
            "06": "#9c009c",
            "7": "#fc7f00",
            "07": "#fc7f00",
            "8": "#ffff00",
            "08": "#ffff00",
            "9": "#00fc00",
            "09": "#00fc00",
            "10": "#009393",
            "11": "#00ffff",
            "12": "#0000fc",
            "13": "#ff00ff",
            "14": "#7f7f7f",
            "15": "#d2d2d2",
            "16": "#470000",
            "17": "#472100",
            "18": "#474700",
            "19": "#324700",
            "20": "#004700",
            "21": "#00472c",
            "22": "#004747",
            "23": "#002747",
            "24": "#000047",
            "25": "#2e0047",
            "26": "#470047",
            "27": "#47002a",
            "28": "#740000",
            "29": "#743a00",
            "30": "#747400",
            "31": "#517400",
            "32": "#007400",
            "33": "#007449",
            "34": "#007474",
            "35": "#004074",
            "36": "#000074",
            "37": "#4b0074",
            "38": "#740074",
            "39": "#740045",
            "40": "#b50000",
            "41": "#b56300",
            "42": "#b5b500",
            "43": "#7db500",
            "44": "#00b500",
            "45": "#00b571",
            "46": "#00b5b5",
            "47": "#0063b5",
            "48": "#0000b5",
            "49": "#7500b5",
            "50": "#b500b5",
            "51": "#b5006b",
            "52": "#ff0000",
            "53": "#ff8c00",
            "54": "#ffff00",
            "55": "#b2ff00",
            "56": "#00ff00",
            "57": "#00ffa0",
            "58": "#00ffff",
            "59": "#008cff",
            "60": "#0000ff",
            "61": "#a500ff",
            "62": "#ff00ff",
            "63": "#ff0098",
            "64": "#ff5959",
            "65": "#ffb459",
            "66": "#ffff71",
            "67": "#cfff60",
            "68": "#6fff6f",
            "69": "#65ffc9",
            "70": "#6dffff",
            "71": "#59b4ff",
            "72": "#5959ff",
            "73": "#c459ff",
            "74": "#ff66ff",
            "75": "#ff59bc",
            "76": "#ff9c9c",
            "77": "#ffd39c",
            "78": "#ffff9c",
            "79": "#e2ff9c",
            "80": "#9cff9c",
            "81": "#9cffdb",
            "82": "#9cffff",
            "83": "#9cd3ff",
            "84": "#9c9cff",
            "85": "#dc9cff",
            "86": "#ff9cff",
            "87": "#ff94d3",
            "88": "#000000",
            "89": "#131313",
            "90": "#282828",
            "91": "#363636",
            "92": "#4d4d4d",
            "93": "#656565",
            "94": "#818181",
            "95": "#9f9f9f",
            "96": "#bcbcbc",
            "97": "#e2e2e2",
            "98": "#ffffff",
        },
    )

    have_formatting = False
    bold = False
    foreground = None
    background = None
    reversed = False
    monospace = False
    italic = False
    strikethrough = False
    underline = False

    for m in re.finditer(
        r"(\x02|\x03(?:([0-9]{1,2})(?:,([0-9]{1,2}))?)?|\x04(?:([0-9A-Fa-f]{6})(?:,([0-9A-Fa-f]{6}))?)?|\x11|\x1D|\x1E|\x1F|\x16|\x0F)?([^\x02\x03\x04\x11\x1D\x1E\x1F\x16\x0F]*)",  # noqa: E501
        input,
    ):
        (ctrl, fg, bg, fghex, bghex, text) = (m.group(1), m.group(2), m.group(3), m.group(4), m.group(5), m.group(6))

        if ctrl:
            have_formatting = True

            if underline:
                formatted.append("</u>")
            if strikethrough:
                formatted.append("</strike>")
            if italic:
                formatted.append("</i>")
            if monospace:
                formatted.append("</code>")
            if color and (foreground is not None or background is not None):
                formatted.append("</font>")
            if bold:
                formatted.append("</b>")

            if ctrl[0] == "\x02":
                bold = not bold
            elif ctrl[0] == "\x03":
                foreground = color_table[fg]
                background = color_table[bg]
            elif ctrl[0] == "\x04":
                foreground = f"#{fghex}"
                background = f"#{bghex}"
            elif ctrl[0] == "\x11":
                monospace = not monospace
            elif ctrl[0] == "\x1D":
                italic = not italic
            elif ctrl[0] == "\x1E":
                strikethrough = not strikethrough
            elif ctrl[0] == "\x1F":
                underline = not underline
            elif ctrl[0] == "\x16":
                reversed = not reversed
            elif ctrl[0] == "\x0F":
                foreground = background = None
                bold = reversed = monospace = italic = strikethrough = underline = False

            if bold:
                formatted.append("<b>")
            if color and (foreground is not None or background is not None):
                formatted.append("<font")
                if not reversed:
                    if foreground is not None:
                        formatted.append(f" data-mx-color='{foreground}'")
                    if background is not None:
                        formatted.append(f" data-mx-bg-color='{background}'")
                else:
                    if background is not None:
                        formatted.append(f" data-mx-color='{background}'")
                    if foreground is not None:
                        formatted.append(f" data-mx-bg-color='{foreground}'")
                formatted.append(">")
            if monospace:
                formatted.append("<code>")
            if italic:
                formatted.append("<i>")
            if strikethrough:
                formatted.append("<strike>")
            if underline:
                formatted.append("<u>")

        if text:
            plain.append(text)

            # escape any existing html in the text
            text = escape(text)

            # create pills
            if pills:
                punct = "?!:;,."

                words = []
                for word in text.split(" "):
                    wlen = len(word)
                    while wlen > 0 and word[wlen - 1] in punct:
                        wlen -= 1

                    word_start = word[:wlen].lower()
                    word_end = word[wlen:]

                    if word_start in pills:
                        mxid, displayname = pills[word_start]
                        words.append(
                            f'<a href="https://matrix.to/#/{escape(mxid)}">{escape(displayname)}</a>{word_end}'
                        )
                    else:
                        words.append(word)

                text = " ".join(words)

            # if the formatted version has a link, we took some pills
            if "<a href" in text:
                have_formatting = True

            formatted.append(text)

    if underline:
        formatted.append("</u>")
    if strikethrough:
        formatted.append("</strike>")
    if italic:
        formatted.append("</i>")
    if monospace:
        formatted.append("</code>")
    if color and (foreground is not None or background is not None):
        formatted.append("</font>")
    if bold:
        formatted.append("</b>")

    return ("".join(plain), "".join(formatted) if have_formatting else None)


def split_long(nick, user, host, target, message):
    out = []

    # this is an easy template to calculate the overhead of the sender and target
    template = f":{nick}!{user}@{host} PRIVMSG {target} :\r\n"
    maxlen = 512 - len(template.encode())
    dots = "..."

    words = []
    for word in message.split(" "):
        words.append(word)
        line = " ".join(words)

        if len(line.encode()) + len(dots) > maxlen:
            words.pop()
            out.append(" ".join(words) + dots)
            words = [dots, word]

    out.append(" ".join(words))

    return out


# generate an edit that follows usual IRC conventions
def line_diff(a, b):
    a = a.split()
    b = b.split()

    pre = None
    post = None
    mlen = min(len(a), len(b))

    for i in range(0, mlen):
        if a[i] != b[i]:
            break

        pre = i + 1

    for i in range(1, mlen + 1):
        if a[-i] != b[-i]:
            break

        post = -i

    rem = a[pre:post]
    add = b[pre:post]

    if len(add) == 0 and len(rem) > 0:
        return "-" + (" ".join(rem))

    if len(rem) == 0 and len(add) > 0:
        return "+" + (" ".join(add))

    if len(add) > 0:
        return "* " + (" ".join(add))

    return None


class PrivateRoom(Room):
    # irc nick of the other party, name for consistency
    name: str
    network: Optional[NetworkRoom]
    network_id: str
    network_name: Optional[str]
    media: List[List[str]]

    max_lines = 0
    use_pastebin = False
    use_reacts = False
    force_forward = False
    prefix_all = False

    commands: CommandManager

    def init(self) -> None:
        self.name = None
        self.network = None
        self.network_id = None
        self.network_name = None  # deprecated
        self.media = []
        self.lazy_members = {}  # allow lazy joining your own ghost for echo

        self.commands = CommandManager()

        if type(self) == PrivateRoom:
            cmd = CommandParser(prog="WHOIS", description="WHOIS the other user")
            self.commands.register(cmd, self.cmd_whois)

        cmd = CommandParser(
            prog="MAXLINES", description="set maximum number of lines per message until truncation or pastebin"
        )
        cmd.add_argument("lines", type=int, nargs="?", help="Number of lines")
        self.commands.register(cmd, self.cmd_maxlines)

        cmd = CommandParser(prog="PASTEBIN", description="enable or disable automatic pastebin of long messages")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable pastebin")
        cmd.add_argument(
            "--disable", dest="enabled", action="store_false", help="Disable pastebin (messages will be truncated)"
        )
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_pastebin)

        cmd = CommandParser(prog="REACTS", description="enable or disable reacting to messages on splits/linking")
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable reacts")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable reacts")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_reacts)

        cmd = CommandParser(
            prog="PREFIXALL", description="prefix all bridged IRC lines with the user's nick, instead of just the first"
        )
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Prefix all lines")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Only prefix first line")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_prefix_all)

        self.mx_register("m.room.message", self.on_mx_message)
        self.mx_register("m.room.redaction", self.on_mx_redaction)

    def from_config(self, config: dict) -> None:
        if "max_lines" in config:
            self.max_lines = config["max_lines"]

        if "use_pastebin" in config:
            self.use_pastebin = config["use_pastebin"]

        if "use_reacts" in config:
            self.use_reacts = config["use_reacts"]

        if "prefix_all" in config:
            self.prefix_all = config["prefix_all"]

        if "name" not in config:
            raise Exception("No name key in config for ChatRoom")

        self.name = config["name"]

        if "network_id" in config:
            self.network_id = config["network_id"]
        if "media" in config:
            self.media = config["media"]

        # only used for migration
        if "network" in config:
            self.network_name = config["network"]

        if self.network_name is None and self.network_id is None:
            raise Exception("No network or network_id key in config for PrivateRoom")

    def to_config(self) -> dict:
        return {
            "name": self.name,
            "network": self.network_name,
            "network_id": self.network_id,
            "media": self.media[:5],
            "max_lines": self.max_lines,
            "use_pastebin": self.use_pastebin,
            "use_reacts": self.use_reacts,
            "prefix_all": self.prefix_all,
        }

    @staticmethod
    def create(network: NetworkRoom, name: str) -> "PrivateRoom":
        logging.debug(f"PrivateRoom.create(network='{network.name}', name='{name}')")
        irc_user_id = network.serv.irc_user_id(network.name, name)
        room = PrivateRoom(
            None,
            network.user_id,
            network.serv,
            [network.user_id, irc_user_id, network.serv.user_id],
            [],
        )
        room.name = name.lower()
        room.network = network
        room.network_id = network.id
        room.network_name = network.name

        room.max_lines = network.serv.config["max_lines"]
        room.use_pastebin = network.serv.config["use_pastebin"]
        room.use_reacts = network.serv.config["use_reacts"]

        asyncio.ensure_future(room._create_mx(name))
        return room

    async def _create_mx(self, displayname) -> None:
        if self.id is None:
            irc_user_id = await self.network.serv.ensure_irc_user_id(self.network.name, displayname, update_cache=False)
            self.id = await self.network.serv.create_room(
                "{} ({})".format(displayname, self.network.name),
                "Private chat with {} on {}".format(displayname, self.network.name),
                [self.network.user_id, irc_user_id],
            )
            self.serv.register_room(self)
            await self.az.intent.user(irc_user_id).ensure_joined(self.id)
            await self.save()
            # start event queue now that we have an id
            self._queue.start()

            # attach to network space
            if self.network.space:
                await self.network.space.attach(self.id)

    def is_valid(self) -> bool:
        if self.network_id is None and self.network_name is None:
            return False

        if self.name is None:
            return False

        if self.user_id is None:
            return False

        if not self.in_room(self.user_id):
            return False

        return True

    def cleanup(self) -> None:
        logging.debug(f"Cleaning up network connected room {self.id}.")

        # cleanup us from network space if we have it
        if self.network and self.network.space:
            asyncio.ensure_future(self.network.space.detach(self.id))

        # cleanup us from network rooms
        if self.network and self.name in self.network.rooms:
            logging.debug(f"... and we are attached to network {self.network.id}, detaching.")
            del self.network.rooms[self.name]

            # if leaving this room invalidated the network, clean it up
            if not self.network.is_valid():
                logging.debug(f"... and we invalidated network {self.network.id} while cleaning up.")
                self.network.serv.unregister_room(self.network.id)
                self.network.cleanup()
                asyncio.ensure_future(self.network.serv.leave_room(self.network.id, self.network.members))

        super().cleanup()

    def send_notice(
        self,
        text: str,
        user_id: Optional[str] = None,
        formatted=None,
        fallback_html: Optional[str] = None,
        forward=False,
    ):
        if (self.force_forward or forward or self.network.forward) and user_id is None:
            self.network.send_notice(text=f"{self.name}: {text}", formatted=formatted, fallback_html=fallback_html)
        else:
            super().send_notice(text=text, user_id=user_id, formatted=formatted, fallback_html=fallback_html)

    def send_notice_html(self, text: str, user_id: Optional[str] = None, forward=False) -> None:
        if (self.force_forward or forward or self.network.forward) and user_id is None:
            self.network.send_notice_html(text=f"{self.name}: {text}")
        else:
            super().send_notice_html(text=text, user_id=user_id)

    def pills(self):
        # if pills are disabled, don't generate any
        if self.network.pills_length < 1:
            return None

        ret = {}
        ignore = list(map(lambda x: x.lower(), self.network.pills_ignore))

        # push our own name first
        lnick = self.network.conn.real_nickname.lower()
        if self.user_id in self.displaynames and len(lnick) >= self.network.pills_length and lnick not in ignore:
            ret[lnick] = (self.user_id, self.displaynames[self.user_id])

        # assuming displayname of a puppet matches nick
        for member in self.members:
            if not member.startswith("@" + self.serv.puppet_prefix) or not member.endswith(":" + self.serv.server_name):
                continue

            if member in self.displaynames:
                nick = self.displaynames[member]
                lnick = nick.lower()
                if len(nick) >= self.network.pills_length and lnick not in ignore:
                    ret[lnick] = (member, nick)

        return ret

    def on_privmsg(self, conn, event) -> None:
        if self.network is None:
            return

        irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)

        (plain, formatted) = parse_irc_formatting(event.arguments[0], self.pills(), self.network.color)

        # ignore relaymsgs by us
        if event.tags:
            for tag in event.tags:
                if tag["key"] == "draft/relaymsg" and tag["value"] == self.network.conn.real_nickname:
                    return

        if event.source.nick == self.network.conn.real_nickname:
            source_irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)

            if self.lazy_members is None:
                self.send_message(f"You said: {plain}", formatted=(f"You said: {formatted}" if formatted else None))
                return
            elif source_irc_user_id not in self.lazy_members:
                # if we are a PM room, remove all other IRC users than the target
                if type(self) == PrivateRoom:
                    target_irc_user_id = self.serv.irc_user_id(self.network.name, self.name)

                    for user_id in self.members:
                        if user_id.startswith("@" + self.serv.puppet_prefix) and user_id != target_irc_user_id:
                            if user_id in self.lazy_members:
                                del self.lazy_members[user_id]
                            self.leave(user_id)

                # add self to lazy members list so it'll echo
                self.lazy_members[source_irc_user_id] = event.source.nick

        if (
            "twitch.tv/membership" in self.network.caps
            and irc_user_id not in self.members
            and irc_user_id not in self.lazy_members
        ):
            self.lazy_members[irc_user_id] = event.source.nick

        self.send_message(
            plain,
            irc_user_id,
            formatted=formatted,
            fallback_html=f"<b>Message from {str(event.source)}</b>: {html.escape(plain)}",
        )

        # lazy update displayname if we detect a change
        if (
            not self.serv.is_user_cached(irc_user_id, event.source.nick)
            and irc_user_id not in (self.lazy_members or {})
            and irc_user_id in self.members
        ):
            asyncio.ensure_future(self.serv.ensure_irc_user_id(self.network.name, event.source.nick))

    def on_privnotice(self, conn, event) -> None:
        if self.network is None:
            return

        (plain, formatted) = parse_irc_formatting(event.arguments[0])

        if event.source.nick == self.network.conn.real_nickname:
            self.send_notice(f"You noticed: {plain}", formatted=(f"You noticed: {formatted}" if formatted else None))
            return

        # if the local user has left this room notify in network
        if self.user_id not in self.members:
            source = self.network.source_text(conn, event)
            self.network.send_notice_html(
                f"Notice from <b>{source}:</b> {formatted if formatted else html.escape(plain)}"
            )
            return

        irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)
        self.send_notice(
            plain,
            irc_user_id,
            formatted=formatted,
            fallback_html=f"<b>Notice from {str(event.source)}</b>: {formatted if formatted else html.escape(plain)}",
        )

    def on_ctcp(self, conn, event) -> None:
        if self.network is None:
            return

        # ignore relaymsgs by us
        if event.tags:
            for tag in event.tags:
                if tag["key"] == "draft/relaymsg" and tag["value"] == self.network.conn.real_nickname:
                    return

        irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)

        command = event.arguments[0].upper()

        if command == "ACTION" and len(event.arguments) > 1:
            (plain, formatted) = parse_irc_formatting(event.arguments[1])

            if event.source.nick == self.network.conn.real_nickname:
                self.send_emote(f"(you) {plain}")
                return

            self.send_emote(
                plain, irc_user_id, fallback_html=f"<b>Emote from {str(event.source)}</b>: {html.escape(plain)}"
            )
        else:
            (plain, formatted) = parse_irc_formatting(" ".join(event.arguments))
            self.send_notice_html(f"<b>{str(event.source)}</b> requested <b>CTCP {html.escape(plain)}</b> (ignored)")

    def on_ctcpreply(self, conn, event) -> None:
        if self.network is None:
            return

        (plain, formatted) = parse_irc_formatting(" ".join(event.arguments))
        self.send_notice_html(f"<b>{str(event.source)}</b> sent <b>CTCP REPLY {html.escape(plain)}</b> (ignored)")

    async def _process_event_content(self, event, prefix, reply_to=None):
        content = event.content

        if content.formatted_body:
            lines = str(await self.parser.parse(content.formatted_body)).split("\n")
        elif content.body:
            lines = content.body.split("\n")
        else:
            logging.warning("_process_event_content called with no usable body")
            return

        # drop all whitespace-only lines
        lines = [x for x in lines if not re.match(r"^\s*$", x)]

        # handle replies
        if reply_to and reply_to.sender != event.sender:
            # resolve displayname
            sender = reply_to.sender
            if sender in self.displaynames:
                sender = self.displaynames[sender]

            # prefix first line with nickname of the reply_to source
            first_line = sender + ": " + lines.pop(0)
            lines.insert(0, first_line)

        messages = []

        for i, line in enumerate(lines):
            # prefix line if needed
            if (i == 0 or self.prefix_all) and prefix and len(prefix) > 0:
                line = prefix + line

                # filter control characters except ZWSP
                line = "".join(c for c in line if unicodedata.category(c)[0] != "C" or c == "\u200B")

            messages += split_long(
                self.network.conn.real_nickname,
                self.network.real_user,
                self.network.real_host,
                self.name,
                line,
            )

        return messages

    async def _send_message(self, event, func, prefix=""):
        # try to find out if this was a reply
        reply_to = None
        if event.content.get_reply_to():
            rel_event = event

            # traverse back all edits
            while rel_event.content.get_edit():
                rel_event = await self.az.intent.get_event(self.id, rel_event.content.get_edit())

            # see if the original is a reply
            if rel_event.content.get_reply_to():
                reply_to = await self.az.intent.get_event(self.id, rel_event.content.get_reply_to())

        if event.content.get_edit():
            messages = await self._process_event_content(event, prefix, reply_to)
            event_id = event.content.relates_to.event_id
            prev_event = self.last_messages[event.sender]
            if prev_event and prev_event.event_id == event_id:
                old_messages = await self._process_event_content(prev_event, prefix, reply_to)

                mlen = max(len(messages), len(old_messages))
                edits = []
                for i in range(0, mlen):
                    try:
                        old_msg = old_messages[i]
                    except IndexError:
                        old_msg = ""
                    try:
                        new_msg = messages[i]
                    except IndexError:
                        new_msg = ""

                    edit = line_diff(old_msg, new_msg)
                    if edit:
                        edits.append(prefix + edit)

                # use edits only if one line was edited
                if len(edits) == 1:
                    messages = edits

                # update last message _content_ to current so re-edits work
                self.last_messages[event.sender].content = event.content
            else:
                # last event was not found so we fall back to full message BUT we can reconstrut enough of it
                self.last_messages[event.sender] = event
        else:
            # keep track of the last message
            self.last_messages[event.sender] = event
            messages = await self._process_event_content(event, prefix, reply_to)

        for i, message in enumerate(messages):
            if self.max_lines > 0 and i == self.max_lines - 1 and len(messages) > self.max_lines:
                if self.use_reacts:
                    self.react(event.event_id, "\u2702")  # scissors

                if self.use_pastebin:
                    content_uri = await self.az.intent.upload_media(
                        "\n".join(messages).encode("utf-8"), mime_type="text/plain; charset=UTF-8"
                    )

                    if self.max_lines == 1:
                        func(
                            self.name,
                            f"{prefix}{self.serv.mxc_to_url(str(content_uri))} (long message, {len(messages)} lines)",
                        )
                    else:
                        func(
                            self.name,
                            f"... long message truncated: {self.serv.mxc_to_url(str(content_uri))} ({len(messages)} lines)",
                        )
                    if self.use_reacts:
                        self.react(event.event_id, "\U0001f4dd")  # memo

                    self.media.append([event.event_id, str(content_uri)])
                    await self.save()
                else:
                    if self.max_lines == 1:
                        # best effort is to send the first line and give up
                        func(self.name, message)
                    else:
                        func(self.name, "... long message truncated")

                return

            func(self.name, message)

        # show number of lines sent to IRC
        if self.use_reacts and self.max_lines == 0 and len(messages) > 1:
            self.react(event.event_id, f"\u2702 {len(messages)} lines")

    async def on_mx_message(self, event) -> None:
        if event.sender != self.user_id:
            return

        if self.network is None or self.network.conn is None or not self.network.conn.connected:
            self.send_notice("Not connected to network.")
            return

        if str(event.content.msgtype) == "m.emote":
            await self._send_message(event, self.network.conn.action)
        elif str(event.content.msgtype) in ["m.image", "m.file", "m.audio", "m.video"]:
            self.network.conn.privmsg(self.name, self.serv.mxc_to_url(event.content.url, event.content.body))
            if self.use_reacts:
                self.react(event.event_id, "\U0001F517")  # link
            self.media.append([event.event_id, event.content.url])
            await self.save()
        elif str(event.content.msgtype) == "m.text":
            # allow commanding the appservice in rooms
            match = re.match(r"^\s*@?([^:,\s]+)[\s:,]*(.+)$", event.content.body)
            if match and match.group(1).lower() == self.serv.registration["sender_localpart"]:
                try:
                    await self.commands.trigger(match.group(2))
                except CommandParserError as e:
                    self.send_notice(str(e))
                finally:
                    return

            await self._send_message(event, self.network.conn.privmsg)

        await self.az.intent.send_receipt(event.room_id, event.event_id)

    async def on_mx_redaction(self, event) -> None:
        for media in self.media:
            if media[0] == event.redacts:
                url = urlparse(media[1])
                if self.serv.synapse_admin:
                    try:
                        await self.az.intent.api.request(
                            Method.POST, SynapseAdminPath.v1.media.quarantine[url.netloc][url.path[1:]]
                        )

                        self.network.send_notice(
                            f"Associated media {media[1]} for redacted event {event.redacts} "
                            + f"in room {self.name} was quarantined."
                        )
                    except Exception:
                        self.network.send_notice(
                            f"Failed to quarantine media! Associated media {media[1]} "
                            + f"for redacted event {event.redacts} in room {self.name} is left available."
                        )
                else:
                    self.network.send_notice(
                        f"No permission to quarantine media! Associated media {media[1]} "
                        + f"for redacted event {event.redacts} in room {self.name} is left available."
                    )
                return

    @connected
    async def cmd_whois(self, args) -> None:
        self.network.conn.whois(f"{self.name} {self.name}")

    async def cmd_maxlines(self, args) -> None:
        if args.lines is not None:
            self.max_lines = args.lines
            await self.save()

        self.send_notice(f"Max lines is {self.max_lines}")

    async def cmd_pastebin(self, args) -> None:
        if args.enabled is not None:
            self.use_pastebin = args.enabled
            await self.save()

        self.send_notice(f"Pastebin is {'enabled' if self.use_pastebin else 'disabled'}")

    async def cmd_reacts(self, args) -> None:
        if args.enabled is not None:
            self.use_reacts = args.enabled
            await self.save()

        self.send_notice(f"Reacts are {'enabled' if self.use_reacts else 'disabled'}")

    async def cmd_prefix_all(self, args) -> None:
        if args.enabled is not None:
            self.prefix_all = args.enabled
            await self.save()

        self.send_notice(f"Prefix all is {'enabled' if self.prefix_all else 'disabled'}")

    async def _attach_hidden_room_internal(self) -> None:
        await self.az.intent.send_state_event(
            self.id,
            EventType.ROOM_JOIN_RULES,
            content=JoinRulesStateEventContent(
                join_rule=JoinRule.RESTRICTED,
                allow=[
                    JoinRestriction(type=JoinRestrictionType.ROOM_MEMBERSHIP, room_id=self.serv.hidden_room.id),
                ],
            ),
        )
        self.hidden_room_id = self.serv.hidden_room.id

    async def _detach_hidden_room_internal(self) -> None:
        await self.az.intent.send_state_event(
            self.id,
            EventType.ROOM_JOIN_RULES,
            content=JoinRulesStateEventContent(join_rule=JoinRule.INVITE),
        )
        self.hidden_room_id = None

    async def _attach_hidden_room(self) -> None:
        if self.hidden_room_id:
            self.send_notice("Room already has a hidden room attached.")
            return
        if not self.serv.hidden_room:
            self.send_notice("Server has no hidden room!")
            return

        logging.debug(f"Attaching room {self.id} to servers hidden room {self.serv.hidden_room.id}.")
        try:
            room_create = await self.az.intent.get_state_event(self.id, EventType.ROOM_CREATE)
            if room_create.room_version in [str(v) for v in range(1, 9)]:
                self.send_notice("Only rooms of version 9 or greater can be attached to a hidden room.")
                self.send_notice("Leave and re-create the room to ensure the correct version.")
                return

            await self._attach_hidden_room_internal()
            self.send_notice("Hidden room attached, invites should now be gone.")
        except MatrixStandardRequestError as e:
            logging.debug("Setting join_rules for hidden room failed.", exc_info=True)
            self.send_notice(f"Failed attaching hidden room: {e.message}")
            self.send_notice("Make sure the room is at least version 9.")
        except Exception:
            logging.exception(f"Failed to attach {self.id} to hidden room {self.serv.hidden_room.id}.")

    async def _detach_hidden_room(self) -> None:
        if not self.hidden_room_id:
            self.send_notice("Room already detached from hidden room.")
            return

        logging.debug(f"Detaching room {self.id} from hidden room {self.hidden_room_id}.")
        try:
            await self._detach_hidden_room_internal()
            self.send_notice("Hidden room detached.")
        except MatrixStandardRequestError as e:
            logging.debug("Setting join_rules for hidden room failed.", exc_info=True)
            self.send_notice(f"Failed detaching hidden room: {e.message}")
        except Exception:
            logging.exception(f"Failed to detach {self.id} from hidden room {self.hidden_room_id}.")

    async def cmd_upgrade(self, args) -> None:
        if args.undo:
            await self._detach_hidden_room()
        else:
            await self._attach_hidden_room()

    async def post_init(self) -> None:
        if self.hidden_room_id and not self.serv.hidden_room:
            logging.debug(
                f"Server has no hidden room, detaching room {self.id} from hidden room {self.hidden_room_id}."
            )
            await self._detach_hidden_room_internal()
        elif self.hidden_room_id and self.hidden_room_id != self.serv.hidden_room.id:
            logging.debug(f"Server has different hidden room, reattaching room {self.id}.")
            await self._attach_hidden_room_internal()



import asyncio
import logging
from typing import List

from mautrix.api import Method
from mautrix.api import Path
from mautrix.types import SpaceChildStateEventContent
from mautrix.types.event.type import EventType

from heisenbridge.room import Room


class NetworkRoom:
    pass


class SpaceRoom(Room):
    # pending rooms to attach during space creation
    pending: List[str]

    def init(self) -> None:
        super().init()

        self.pending = []

    def is_valid(self) -> bool:
        # we need to know our network
        if self.network_id is None:
            return False

        # we are valid as long as our user is in the room
        if not self.in_room(self.user_id):
            return False

        return True

    @staticmethod
    def create(network: "NetworkRoom", initial_rooms: List[str]) -> "SpaceRoom":
        logging.debug(f"SpaceRoom.create(network='{network.id}' ({network.name}))")

        room = SpaceRoom(
            None,
            network.user_id,
            network.serv,
            [network.user_id, network.serv.user_id],
            [],
        )
        room.name = network.name
        room.network = network  # only used in create_finalize
        room.network_id = network.id
        room.pending += initial_rooms
        return room

    async def create_finalize(self) -> None:
        resp = await self.az.intent.api.request(
            Method.POST,
            Path.v3.createRoom,
            {
                "creation_content": {
                    "type": "m.space",
                },
                "visibility": "private",
                "name": self.network.name,
                "topic": f"Network space for {self.network.name}",
                "invite": [self.network.user_id],
                "is_direct": False,
                "initial_state": [
                    {
                        "type": "m.space.child",
                        "state_key": self.network.id,
                        "content": {"via": [self.network.serv.server_name]},
                    }
                ],
                "power_level_content_override": {
                    "events_default": 100,
                    "users_default": 0,
                    "invite": 100,
                    "kick": 100,
                    "redact": 100,
                    "ban": 100,
                    "events": {
                        "m.room.name": 0,
                        "m.room.avatar": 0,  # these work as long as rooms are private
                    },
                },
            },
        )

        self.id = resp["room_id"]
        self.serv.register_room(self)
        await self.save()

        # attach all pending rooms
        rooms = self.pending
        self.pending = []

        for room_id in rooms:
            await self.attach(room_id)

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "network_id" in config:
            self.network_id = config["network_id"]

    def to_config(self) -> dict:
        return {
            **(super().to_config()),
            "network_id": self.network_id,
        }

    def cleanup(self) -> None:
        try:
            network = self.serv._rooms[self.network_id]

            if network.space == self:
                network.space = None
                network.space_id = None
                asyncio.ensure_future(network.save())
                logging.debug(f"Space {self.id} cleaned up from network {network.id}")
            else:
                logging.debug(f"Space room cleaned up as a duplicate for network {network.id}, probably fine.")
        except KeyError:
            logging.debug(f"Space room cleaned up with missing network {self.network_id}, probably fine.")

        super().cleanup()

    async def attach(self, room_id) -> None:
        # if we are attached between space request and creation just add to pending list
        if self.id is None:
            logging.debug(f"Queuing room {room_id} attachment to pending space.")
            self.pending.append(room_id)
            return

        logging.debug(f"Attaching room {room_id} to space {self.id}.")
        await self.az.intent.send_state_event(
            self.id,
            EventType.SPACE_CHILD,
            state_key=room_id,
            content=SpaceChildStateEventContent(via=[self.serv.server_name]),
        )

    async def detach(self, room_id) -> None:
        if self.id is not None:
            logging.debug(f"Detaching room {room_id} from space {self.id}.")
            await self.az.intent.send_state_event(
                self.id, EventType.SPACE_CHILD, state_key=room_id, content=SpaceChildStateEventContent()
            )
        elif room_id in self.pending:
            logging.debug(f"Removing {room_id} from space {self.id} pending queue.")
            self.pending.remove(room_id)

    async def post_init(self) -> None:
        try:
            network = self.serv._rooms[self.network_id]
            if network.space is not None:
                logging.warn(
                    f"Network room {network.id} already has space {network.space.id} but I'm {self.id}, we are dangling."
                )
                return

            network.space = self
            logging.debug(f"Space {self.id} attached to network {network.id}")
        except KeyError:
            logging.debug(f"Network room {self.network_id} was not found for space {self.id}, we are dangling.")
            self.network_id = None



import asyncio
import logging
import re
import socket

from heisenbridge.network_room import NetworkRoom


class Identd:
    async def handle(self, reader, writer):
        try:
            data = await asyncio.wait_for(reader.readuntil(b"\r\n"), 10)
            query = data.decode()

            m = re.match(r"^(\d+)\s*,\s*(\d+)", query)
            if m:
                req_addr, req_port, *_ = writer.get_extra_info("peername")

                src_port = int(m.group(1))
                dst_port = int(m.group(2))

                response = f"{src_port}, {dst_port} : ERROR : NO-USER\r\n"

                logging.debug(f"Remote {req_addr} wants to know who is {src_port} connected to {dst_port}")

                """
                This is a hack to workaround the issue where asyncio create_connection has not returned before
                identd is already requested.

                Proper fix would be to use our own sock that has been pre-bound but that's quite a bit of work
                for very little gain.
                """
                await asyncio.sleep(0.1)

                for room in self.serv.find_rooms(NetworkRoom):
                    if not room.conn or not room.conn.connected:
                        continue

                    remote_addr, remote_port, *_ = room.conn.transport.get_extra_info("peername") or ("", "")
                    local_addr, local_port, *_ = room.conn.transport.get_extra_info("sockname") or ("", "")

                    if remote_port == dst_port and local_port == src_port:
                        response = f"{src_port}, {dst_port} : USERID : UNIX : {room.get_ident()}\r\n"
                        break

                logging.debug(f"Responding with: {response}")
                writer.write(response.encode())
                await writer.drain()
        except Exception:
            logging.debug("Identd request threw exception, ignored")
        finally:
            writer.close()

    async def start_listening(self, serv, port):
        self.serv = serv

        # XXX: this only works if dual stack is enabled which usually is
        if socket.has_ipv6:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            sock.bind(("::", port))
            self.server = await asyncio.start_server(self.handle, sock=sock, limit=128)
        else:
            self.server = await asyncio.start_server(self.handle, "0.0.0.0", port, limit=128)
